# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from collections.abc import Iterable
from copy import deepcopy
from dataclasses import fields
from functools import singledispatchmethod
from logging import Logger, getLogger
from typing import Optional, Union

import numpy as np
from sympy import Symbol

from braket.ir.openqasm.program_v1 import io_type

from ._helpers.arrays import (
    convert_range_def_to_range,
    create_empty_array,
    flatten_indices,
    get_elements,
    get_type_width,
)
from ._helpers.casting import (
    LiteralType,
    cast_to,
    get_identifier_name,
    is_literal,
    wrap_value_into_literal,
)
from ._helpers.functions import (
    builtin_constants,
    builtin_functions,
    evaluate_binary_expression,
    evaluate_unary_expression,
    get_operator_of_assignment_operator,
)
from ._helpers.quantum import (
    convert_phase_to_gate,
    get_ctrl_modifiers,
    get_pow_modifiers,
    invert_phase,
    is_controlled,
    is_inverted,
    modify_body,
)
from .circuit import Circuit
from .parser.openqasm_ast import (
    AccessControl,
    ArrayLiteral,
    ArrayReferenceType,
    ArrayType,
    AssignmentOperator,
    BinaryExpression,
    BitstringLiteral,
    BitType,
    BooleanLiteral,
    BranchingStatement,
    Cast,
    ClassicalArgument,
    ClassicalAssignment,
    ClassicalDeclaration,
    ConstantDeclaration,
    DiscreteSet,
    FloatLiteral,
    ForInLoop,
    FunctionCall,
    GateModifierName,
    Identifier,
    Include,
    IndexedIdentifier,
    IndexExpression,
    IntegerLiteral,
    IODeclaration,
    IOKeyword,
    Pragma,
    Program,
    QASMNode,
    QuantumGate,
    QuantumGateDefinition,
    QuantumGateModifier,
    QuantumMeasurement,
    QuantumMeasurementStatement,
    QuantumPhase,
    QuantumReset,
    QuantumStatement,
    QubitDeclaration,
    RangeDefinition,
    ReturnStatement,
    SizeOf,
    SubroutineDefinition,
    SymbolLiteral,
    UnaryExpression,
    WhileLoop,
)
from .parser.openqasm_parser import parse
from .program_context import AbstractProgramContext, ProgramContext


class Interpreter:
    """
    The interpreter is responsible for visiting the AST of an OpenQASM program, as created
    by the parser, and building a braket.default_simulator.openqasm.circuit.Circuit to hand
    off to a simulator e.g. braket.default_simulator.state_vector_simulator.StateVectorSimulator.

    The interpreter keeps track of all state using a ProgramContext object. The main entry point
    is build_circuit(), which returns the built circuit. An alternative entry poitn, run() returns
    the ProgramContext object, which can be used for debugging or other customizability.
    """

    def __init__(
        self, context: Optional[AbstractProgramContext] = None, logger: Optional[Logger] = None
    ):
        # context keeps track of all state
        self.context = context or ProgramContext()
        self.logger = logger or getLogger(__name__)
        self._uses_advanced_language_features = False

    def build_circuit(
        self, source: str, inputs: Optional[dict[str, io_type]] = None, is_file: bool = False
    ) -> Circuit:
        """Interpret an OpenQASM program and build a Circuit IR."""
        return self.run(source, inputs, is_file).circuit

    def run(
        self, source: str, inputs: Optional[dict[str, io_type]] = None, is_file: bool = False
    ) -> ProgramContext:
        """Interpret an OpenQASM program and return the program state"""
        if inputs:
            self.context.load_inputs(inputs)

        if is_file:
            with open(source, encoding="utf-8") as f:
                source = f.read()

        program = parse(source)
        self._uses_advanced_language_features = False
        self.visit(program)
        if self._uses_advanced_language_features:
            self.logger.warning(
                "This program uses OpenQASM language features that may "
                "not be supported on QPUs or on-demand simulators."
            )
        return self.context

    @singledispatchmethod
    def visit(self, node: Union[QASMNode, list[QASMNode]]) -> Optional[QASMNode]:
        """Generic visit function for an AST node"""
        if node is None:
            return
        if not isinstance(node, QASMNode):
            return node
        for field in fields(node):
            value = getattr(node, field.name)
            setattr(node, field.name, self.visit(value))
        return node

    @visit.register
    def _(self, node_list: list) -> list[QASMNode]:
        """Generic visit function for a list of AST nodes"""
        return [n for n in [self.visit(node) for node in node_list] if n is not None]

    @visit.register
    def _(self, node: Program) -> None:
        self.visit(node.statements)

    @visit.register
    def _(self, node: ClassicalDeclaration) -> None:
        node_type = self.visit(node.type)
        if node.init_expression is not None:
            init_expression = self.visit(node.init_expression)
            init_value = cast_to(node.type, init_expression)
        elif isinstance(node_type, ArrayType):
            init_value = create_empty_array(node_type.dimensions)
        elif isinstance(node_type, BitType) and node_type.size:
            init_value = create_empty_array([node_type.size])
        else:
            init_value = None
        self.context.declare_variable(node.identifier.name, node_type, init_value)

    @visit.register
    def _(self, node: IODeclaration) -> None:
        if node.io_identifier == IOKeyword.output:
            raise NotImplementedError("Output not supported")
        else:  # IOKeyword.input:
            if node.identifier.name not in self.context.inputs:
                # previously raised a NameError
                init_value = wrap_value_into_literal(Symbol(node.identifier.name))
                node_type = SymbolLiteral
            else:
                init_value = wrap_value_into_literal(self.context.inputs[node.identifier.name])
                node_type = node.type
            declaration = ClassicalDeclaration(node_type, node.identifier, init_value)
            self.visit(declaration)

    @visit.register
    def _(self, node: ConstantDeclaration) -> None:
        self._uses_advanced_language_features = True
        node_type = self.visit(node.type)
        init_expression = self.visit(node.init_expression)
        init_value = cast_to(node.type, init_expression)
        self.context.declare_variable(node.identifier.name, node_type, init_value, const=True)

    @visit.register
    def _(self, node: BinaryExpression) -> Union[BinaryExpression, LiteralType]:
        lhs = self.visit(node.lhs)
        rhs = self.visit(node.rhs)
        if is_literal(lhs) and is_literal(rhs):
            return evaluate_binary_expression(lhs, rhs, node.op)
        else:
            return BinaryExpression(node.op, lhs, rhs)

    @visit.register
    def _(self, node: UnaryExpression) -> Union[UnaryExpression, LiteralType]:
        expression = self.visit(node.expression)
        if is_literal(expression):
            return evaluate_unary_expression(expression, node.op)
        else:
            return UnaryExpression(node.op, expression)

    @visit.register
    def _(self, node: Cast) -> LiteralType:
        return cast_to(node.type, self.visit(node.argument))

    @visit.register(BooleanLiteral)
    @visit.register(IntegerLiteral)
    @visit.register(FloatLiteral)
    def _(self, node: LiteralType) -> LiteralType:
        return node

    @visit.register
    def _(self, node: Identifier) -> LiteralType:
        if node.name.startswith("$"):
            return node
        if node.name in builtin_constants:
            return builtin_constants[node.name]
        if not self.context.is_initialized(node.name):
            raise NameError(f"Identifier '{node.name}' is not initialized.")
        return self.context.get_value_by_identifier(node)

    @visit.register
    def _(self, node: QubitDeclaration) -> None:
        size = self.visit(node.size).value if node.size else 1
        self.context.add_qubits(node.qubit.name, size)

    @visit.register
    def _(self, node: QuantumReset) -> None:
        raise NotImplementedError("Reset not supported")

    @visit.register
    def _(self, node: IndexedIdentifier) -> Union[IndexedIdentifier, LiteralType]:
        """Returns an identifier for qubits, value for classical identifier"""
        name = node.name
        indices = []
        for index in node.indices:
            if isinstance(index, DiscreteSet):
                self._uses_advanced_language_features = True
                indices.append(index)
            else:
                for element in index:
                    if isinstance(element, RangeDefinition):
                        self._uses_advanced_language_features = True
                    element = self.visit(element)
                    indices.append([element])
        updated = IndexedIdentifier(name, indices)
        if name.name not in self.context.qubit_mapping:
            return self.context.get_value_by_identifier(updated)
        return updated

    @visit.register
    def _(self, node: RangeDefinition) -> RangeDefinition:
        self._uses_advanced_language_features = True
        start = self.visit(node.start) if node.start else None
        end = self.visit(node.end)
        step = self.visit(node.step) if node.step else None
        return RangeDefinition(start, end, step)

    @visit.register
    def _(self, node: IndexExpression) -> Union[IndexedIdentifier, ArrayLiteral]:
        """Returns an identifier for qubits, values for classical identifier"""
        type_width = None
        index = self.visit(node.index)
        if isinstance(node.collection, Identifier):
            # indexed QuantumArgument
            if isinstance(self.context.get_type(node.collection.name), type(Identifier)):
                return IndexedIdentifier(node.collection, [index])
            var_type = self.context.get_type(get_identifier_name(node.collection))
            type_width = get_type_width(var_type)
        collection = self.visit(node.collection)
        return get_elements(collection, index, type_width)

    @visit.register
    def _(self, node: QuantumGateDefinition) -> None:
        self._uses_advanced_language_features = True
        with self.context.enter_scope():
            for qubit in node.qubits:
                self.context.declare_qubit_alias(qubit.name, qubit)

            for param in node.arguments:
                self.context.declare_variable(param.name, Identifier, param)

            node.body = self.inline_gate_def_body(node.body)
        self.context.add_gate(node.name.name, node)

    def inline_gate_def_body(self, body: list[QuantumStatement]) -> list[QuantumStatement]:
        inlined_body = []
        for statement in body:
            if isinstance(statement, QuantumPhase):
                statement.argument = self.visit(statement.argument)
                statement.modifiers = self.visit(statement.modifiers)
                if is_inverted(statement):
                    statement = invert_phase(statement)
                if is_controlled(statement):
                    statement = convert_phase_to_gate(statement)
                # statement is a quantum phase instruction
                else:
                    inlined_body.append(statement)
            # this includes converted phase instructions
            if isinstance(statement, QuantumGate):
                gate_name = statement.name.name
                statement.arguments = self.visit(statement.arguments)
                statement.modifiers = self.visit(statement.modifiers)
                statement.qubits = self.visit(statement.qubits)
                if self.context.is_builtin_gate(gate_name):
                    inlined_body.append(statement)
                else:
                    with self.context.enter_scope():
                        gate_def = self.context.get_gate_definition(gate_name)
                        ctrl_modifiers = get_ctrl_modifiers(statement.modifiers)
                        pow_modifiers = get_pow_modifiers(statement.modifiers)
                        num_ctrl = sum(mod.argument.value for mod in ctrl_modifiers)
                        ctrl_qubits = statement.qubits[:num_ctrl]
                        gate_qubits = statement.qubits[num_ctrl:]

                        for qubit_called, qubit_defined in zip(gate_qubits, gate_def.qubits):
                            self.context.declare_qubit_alias(qubit_defined.name, qubit_called)

                        for param_called, param_defined in zip(
                            statement.arguments, gate_def.arguments
                        ):
                            self.context.declare_variable(
                                param_defined.name, Identifier, param_called
                            )

                        inlined_copy = self.inline_gate_def_body(deepcopy(gate_def.body))

                    inlined_body += modify_body(
                        inlined_copy,
                        is_inverted(statement),
                        ctrl_modifiers,
                        ctrl_qubits,
                        pow_modifiers,
                    )
        return inlined_body

    @visit.register
    def _(self, node: QuantumGate) -> None:
        gate_name = node.name.name
        arguments = self.visit(node.arguments)
        modifiers = self.visit(node.modifiers)
        if self.context.in_global_scope and modifiers:
            self._uses_advanced_language_features = True

        qubits = []
        for qubit in node.qubits:
            if isinstance(qubit, Identifier):
                qubits.append(self.visit(qubit))
            else:  # IndexedIdentifier
                dereffed_name = self.visit(qubit.name)
                simplified_indices = self.visit(qubit.indices)
                qubits.append(IndexedIdentifier(dereffed_name, simplified_indices))

        qubit_lengths = np.array(
            [self.context.qubit_mapping.get_qubit_size(qubit) for qubit in qubits]
        )
        register_lengths = qubit_lengths[qubit_lengths > 1]
        if register_lengths.size:
            reg_length = register_lengths[0]
            if not np.all(register_lengths == reg_length):
                raise ValueError("Qubit registers must all be the same length.")

            for i in range(reg_length):
                indexed_qubits = deepcopy(qubits)
                for j, qubit_length in enumerate(qubit_lengths):
                    if qubit_length > 1:
                        if isinstance(indexed_qubits[j], Identifier):
                            indexed_qubits[j] = IndexedIdentifier(
                                indexed_qubits[j], [[IntegerLiteral(i)]]
                            )
                        else:
                            indexed_qubits[j].indices.append([IntegerLiteral(i)])
                gate_call = QuantumGate(
                    modifiers,
                    node.name,
                    arguments,
                    indexed_qubits,
                )
                self.visit(gate_call)
            return

        if self.context.is_builtin_gate(gate_name):
            # to simplify indices
            qubits = self.visit(qubits)
            self.handle_builtin_gate(
                gate_name,
                arguments,
                qubits,
                modifiers,
            )
        else:
            with self.context.enter_scope():
                gate_def = self.context.get_gate_definition(gate_name)

                ctrl_modifiers = get_ctrl_modifiers(modifiers)
                pow_modifiers = get_pow_modifiers(modifiers)
                num_ctrl = sum(mod.argument.value for mod in ctrl_modifiers)
                ctrl_qubits = qubits[:num_ctrl]
                gate_qubits = qubits[num_ctrl:]

                modified_gate_body = modify_body(
                    deepcopy(gate_def.body),
                    is_inverted(node),
                    ctrl_modifiers,
                    ctrl_qubits,
                    pow_modifiers,
                )

                for qubit_called, qubit_defined in zip(gate_qubits, gate_def.qubits):
                    self.context.declare_qubit_alias(qubit_defined.name, qubit_called)

                for param_called, param_defined in zip(arguments, gate_def.arguments):
                    self.context.declare_variable(param_defined.name, FloatLiteral, param_called)

                for statement in deepcopy(modified_gate_body):
                    if isinstance(statement, QuantumGate):
                        self.visit(statement)
                    else:  # QuantumPhase
                        phase = self.visit(statement.argument)
                        self.handle_phase(phase, qubits)

    @visit.register
    def _(self, node: QuantumPhase) -> None:
        node.argument = self.visit(node.argument)
        node.modifiers = self.visit(node.modifiers)
        if is_inverted(node):
            node = invert_phase(node)
        if is_controlled(node):
            node = convert_phase_to_gate(node)
            self.visit(node)
        else:
            self.handle_phase(node.argument)

    @visit.register
    def _(self, node: QuantumGateModifier) -> QuantumGateModifier:
        if node.modifier in (GateModifierName.ctrl, GateModifierName.negctrl):
            if node.argument is None:
                node.argument = IntegerLiteral(1)
            else:
                node.argument = self.visit(node.argument)
        elif node.modifier == GateModifierName.pow:
            node.argument = self.visit(node.argument)
        return node

    @visit.register
    def _(self, node: QuantumMeasurement) -> None:
        qubits = self.context.get_qubits(self.visit(node.qubit))
        return qubits

    @visit.register
    def _(self, node: QuantumMeasurementStatement) -> None:
        """The measure is performed but the assignment is ignored"""
        qubits = self.visit(node.measure)
        targets = []
        if node.target:
            if isinstance(node.target, IndexedIdentifier):
                indices = flatten_indices(node.target.indices)
                if len(node.target.indices) != 1:
                    raise ValueError(
                        "Multi-Dimensional indexing not supported for classical registers."
                    )
                elem = indices[0]
                if isinstance(elem, DiscreteSet):
                    self._uses_advanced_language_features = True
                    target_indices = [self.visit(val).value for val in elem.values]
                    targets.extend(target_indices)
                elif isinstance(elem, RangeDefinition):
                    self._uses_advanced_language_features = True
                    target_indices = convert_range_def_to_range(self.visit(elem))
                    targets.extend(target_indices)
                else:
                    target_idx = elem.value
                    targets.append(target_idx)

        if not len(targets):
            targets = None

        if targets and len(targets) != len(qubits):
            raise ValueError(
                f"Number of qubits ({len(qubits)}) does not match number of provided classical targets ({len(targets)})"
            )
        self.context.add_measure(qubits, targets)

    @visit.register
    def _(self, node: ClassicalAssignment) -> None:
        lvalue_name = get_identifier_name(node.lvalue)
        if self.context.get_const(lvalue_name):
            raise TypeError(f"Cannot update const value {lvalue_name}")
        if node.op == getattr(AssignmentOperator, "="):
            rvalue = self.visit(node.rvalue)
        else:
            op = get_operator_of_assignment_operator(node.op)
            binary_expression = BinaryExpression(op, node.lvalue, node.rvalue)
            rvalue = self.visit(binary_expression)
        lvalue = node.lvalue
        if isinstance(lvalue, IndexedIdentifier):
            lvalue.indices = self.visit(lvalue.indices)
        elif isinstance(rvalue, SymbolLiteral):
            pass
        else:
            rvalue = cast_to(self.context.get_type(lvalue.name), rvalue)
        self.context.update_value(lvalue, rvalue)

    @visit.register
    def _(self, node: BitstringLiteral) -> ArrayLiteral:
        return cast_to(BitType(IntegerLiteral(node.width)), node)

    @visit.register
    def _(self, node: BranchingStatement) -> None:
        self._uses_advanced_language_features = True
        condition = cast_to(BooleanLiteral, self.visit(node.condition))
        block = node.if_block if condition.value else node.else_block
        for statement in block:
            self.visit(statement)

    @visit.register
    def _(self, node: ForInLoop) -> None:
        self._uses_advanced_language_features = True
        index = self.visit(node.set_declaration)
        if isinstance(index, RangeDefinition):
            index_values = [IntegerLiteral(x) for x in convert_range_def_to_range(index)]
        # DiscreteSet
        else:
            index_values = index.values
        block = node.block
        for i in index_values:
            block_copy = deepcopy(block)
            with self.context.enter_scope():
                self.context.declare_variable(node.identifier.name, node.type, i)
                self.visit(block_copy)

    @visit.register
    def _(self, node: WhileLoop) -> None:
        self._uses_advanced_language_features = True
        while cast_to(BooleanLiteral, self.visit(deepcopy(node.while_condition))).value:
            self.visit(deepcopy(node.block))

    @visit.register
    def _(self, node: Include) -> None:
        self._uses_advanced_language_features = True
        with open(node.filename, encoding="utf-8") as f:
            included = f.read()
            parsed = parse(included)
            self.visit(parsed)

    @visit.register
    def _(self, node: Pragma) -> None:
        parsed = self.context.parse_pragma(node.command)
        if node.command.startswith("braket result"):
            if not parsed:
                raise TypeError(f"Result type {node.command.split()[2]} is not supported.")
            self.context.add_result(parsed)
        elif node.command.startswith("braket unitary"):
            unitary, target = parsed
            self.context.add_custom_unitary(unitary, target)
        elif node.command.startswith("braket noise kraus"):
            matrices, target = parsed
            self.context.add_kraus_instruction(matrices, target)
        elif node.command.startswith("braket noise"):
            noise_instruction, target, probabilities = parsed
            self.context.add_noise_instruction(noise_instruction, target, probabilities)
        elif node.command.startswith("braket verbatim"):
            pass
        else:
            raise NotImplementedError(f"Pragma '{node.command}' is not supported")

    @visit.register
    def _(self, node: SubroutineDefinition) -> None:
        self._uses_advanced_language_features = True
        # todo: explicitly handle references to existing variables
        # either by throwing an error or evaluating the closure.
        # currently, the implementation does not consider the values
        # of current-scope variables used inside of the function
        # at the time of function definition, and relies on their values
        # at the time of execution. This is incorrect, but currently an
        # edge case and known limitation. More effort can be invested here
        # if this functionality is prioritized.
        self.context.add_subroutine(node.name.name, node)

    @visit.register
    def _(self, node: FunctionCall) -> Optional[QASMNode]:
        self._uses_advanced_language_features = True
        function_name = node.name.name
        arguments = self.visit(node.arguments)
        if function_name in builtin_functions:
            return builtin_functions[function_name](*arguments)
        function_def = self.context.get_subroutine_definition(function_name)
        with self.context.enter_scope():
            for arg_passed, arg_defined in zip(arguments, function_def.arguments):
                if isinstance(arg_defined, ClassicalArgument):
                    arg_name = arg_defined.name.name
                    arg_type = arg_defined.type
                    arg_const = arg_defined.access == AccessControl.const
                    arg_value = deepcopy(arg_passed)

                    self.context.declare_variable(arg_name, arg_type, arg_value, arg_const)

                else:  # QuantumArgument
                    qubit_name = get_identifier_name(arg_defined.name)
                    self.context.declare_qubit_alias(qubit_name, arg_passed)

            return_value = None
            for statement in deepcopy(function_def.body):
                visited = self.visit(statement)
                if isinstance(statement, ReturnStatement):
                    return_value = visited
                    break

            for arg_passed, arg_defined in zip(node.arguments, function_def.arguments):
                if isinstance(arg_defined, ClassicalArgument):
                    if isinstance(arg_defined.type, ArrayReferenceType):
                        if isinstance(arg_passed, IndexExpression):
                            identifier = IndexedIdentifier(
                                arg_passed.collection, [arg_passed.index]
                            )
                            identifier.indices = self.visit(identifier.indices)
                        else:
                            identifier = arg_passed
                        reference_value = self.context.get_value(arg_defined.name.name)
                        self.context.update_value(identifier, reference_value)

            return return_value

    @visit.register
    def _(self, node: ReturnStatement) -> Optional[QASMNode]:
        self._uses_advanced_language_features = True
        return self.visit(node.expression)

    @visit.register
    def _(self, node: SizeOf) -> IntegerLiteral:
        self._uses_advanced_language_features = True
        target = self.visit(node.target)
        index = self.visit(node.index)
        return builtin_functions["sizeof"](target, index)

    def handle_builtin_gate(
        self,
        gate_name: str,
        arguments: list[FloatLiteral],
        qubits: list[Union[Identifier, IndexedIdentifier]],
        modifiers: list[QuantumGateModifier],
    ) -> None:
        """Add unitary operation to the circuit"""
        self.context.add_builtin_gate(
            gate_name,
            arguments,
            qubits,
            modifiers,
        )

    def handle_phase(self, phase: FloatLiteral, qubits: Optional[Iterable[int]] = None) -> None:
        """Add quantum phase operation to the circuit"""
        self.context.add_phase(phase, qubits)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from typing import Union

from ..parser.openqasm_ast import (
    FloatLiteral,
    GateModifierName,
    Identifier,
    IntegerLiteral,
    QuantumGate,
    QuantumGateModifier,
    QuantumPhase,
    QuantumStatement,
)


def invert_phase(phase: QuantumPhase) -> QuantumPhase:
    """Invert a quantum phase"""
    new_modifiers = [mod for mod in phase.modifiers if mod.modifier != GateModifierName.inv]
    new_param = FloatLiteral(-phase.argument.value)
    return QuantumPhase(new_modifiers, new_param, phase.qubits)


def is_inverted(quantum_op: Union[QuantumGate, QuantumPhase]) -> bool:
    """
    Tell whether a gate with modifiers is inverted, or if the inverse modifiers
    cancel out. Since inv @ ctrl U == ctrl @ inv U, we can accomplish this by
    only counting the inverse modifiers.
    """
    inv_modifier = QuantumGateModifier(GateModifierName.inv, None)
    num_inv_modifiers = quantum_op.modifiers.count(inv_modifier)
    return bool(num_inv_modifiers % 2)


def is_controlled(phase: QuantumPhase) -> bool:
    """
    Returns whether a quantum phase has any control modifiers. If it does, then
    it will be transformed by the interpreter into a controlled global phase gate.
    """
    for mod in phase.modifiers:
        if mod.modifier in (GateModifierName.ctrl, GateModifierName.negctrl):
            return True
    return False


def convert_phase_to_gate(controlled_phase: QuantumPhase) -> QuantumGate:
    """Convert a controlled quantum phase into a quantum gate"""
    ctrl_modifiers = get_ctrl_modifiers(controlled_phase.modifiers)
    first_ctrl_modifier = ctrl_modifiers[-1]
    if first_ctrl_modifier.modifier == GateModifierName.negctrl:
        raise ValueError("negctrl modifier undefined for gphase operation")
    if first_ctrl_modifier.argument.value == 1:
        ctrl_modifiers.pop()
    else:
        ctrl_modifiers[-1].argument.value -= 1
    return QuantumGate(
        ctrl_modifiers,
        Identifier("U"),
        [
            IntegerLiteral(0),
            IntegerLiteral(0),
            controlled_phase.argument,
        ],
        controlled_phase.qubits,
    )


def get_ctrl_modifiers(modifiers: list[QuantumGateModifier]) -> list[QuantumGateModifier]:
    """Get the control modifiers from a list of quantum gate modifiers"""
    return [
        mod
        for mod in modifiers
        if mod.modifier in (GateModifierName.ctrl, GateModifierName.negctrl)
    ]


def get_pow_modifiers(modifiers: list[QuantumGateModifier]) -> list[QuantumGateModifier]:
    """Get the power modifiers from a list of quantum gate modifiers"""
    return [mod for mod in modifiers if mod.modifier == GateModifierName.pow]


def modify_body(
    body: list[QuantumStatement],
    do_invert: bool,
    ctrl_modifiers: list[QuantumGateModifier],
    ctrl_qubits: list[Identifier],
    pow_modifiers: list[QuantumGateModifier],
) -> list[QuantumStatement]:
    """Apply modifiers information to the definition body of a quantum gate"""
    if do_invert:
        body = list(reversed(body))
        for s in body:
            s.modifiers.insert(0, QuantumGateModifier(GateModifierName.inv, None))
    for s in body:
        if isinstance(s, QuantumGate) or is_controlled(s):
            s.modifiers = ctrl_modifiers + pow_modifiers + s.modifiers
            s.qubits = ctrl_qubits + s.qubits
    return body



# Generated from BraketPragmasParser.g4 by ANTLR 4.9.2
from antlr4 import ParseTreeVisitor

if __name__ is not None and "." in __name__:
    from .BraketPragmasParser import BraketPragmasParser
else:
    from BraketPragmasParser import BraketPragmasParser

# This class defines a complete generic visitor for a parse tree produced by BraketPragmasParser.


class BraketPragmasParserVisitor(ParseTreeVisitor):
    # Visit a parse tree produced by BraketPragmasParser#braketPragma.
    def visitBraketPragma(self, ctx: BraketPragmasParser.BraketPragmaContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#braketUnitaryPragma.
    def visitBraketUnitaryPragma(self, ctx: BraketPragmasParser.BraketUnitaryPragmaContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#braketVerbatimPragma.
    def visitBraketVerbatimPragma(self, ctx: BraketPragmasParser.BraketVerbatimPragmaContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#twoDimMatrix.
    def visitTwoDimMatrix(self, ctx: BraketPragmasParser.TwoDimMatrixContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#row.
    def visitRow(self, ctx: BraketPragmasParser.RowContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#braketResultPragma.
    def visitBraketResultPragma(self, ctx: BraketPragmasParser.BraketResultPragmaContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#resultType.
    def visitResultType(self, ctx: BraketPragmasParser.ResultTypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#noArgResultType.
    def visitNoArgResultType(self, ctx: BraketPragmasParser.NoArgResultTypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#noArgResultTypeName.
    def visitNoArgResultTypeName(self, ctx: BraketPragmasParser.NoArgResultTypeNameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#optionalMultiTargetResultType.
    def visitOptionalMultiTargetResultType(
        self, ctx: BraketPragmasParser.OptionalMultiTargetResultTypeContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#optionalMultiTargetResultTypeName.
    def visitOptionalMultiTargetResultTypeName(
        self, ctx: BraketPragmasParser.OptionalMultiTargetResultTypeNameContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#MultiTargetIdentifiers.
    def visitMultiTargetIdentifiers(self, ctx: BraketPragmasParser.MultiTargetIdentifiersContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#MultiTargetAll.
    def visitMultiTargetAll(self, ctx: BraketPragmasParser.MultiTargetAllContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#multiStateResultType.
    def visitMultiStateResultType(self, ctx: BraketPragmasParser.MultiStateResultTypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#multiStateResultTypeName.
    def visitMultiStateResultTypeName(
        self, ctx: BraketPragmasParser.MultiStateResultTypeNameContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#multiState.
    def visitMultiState(self, ctx: BraketPragmasParser.MultiStateContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#observableResultType.
    def visitObservableResultType(self, ctx: BraketPragmasParser.ObservableResultTypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#observable.
    def visitObservable(self, ctx: BraketPragmasParser.ObservableContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#StandardObservableIdentifier.
    def visitStandardObservableIdentifier(
        self, ctx: BraketPragmasParser.StandardObservableIdentifierContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#StandardObservableAll.
    def visitStandardObservableAll(self, ctx: BraketPragmasParser.StandardObservableAllContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#tensorProductObservable.
    def visitTensorProductObservable(self, ctx: BraketPragmasParser.TensorProductObservableContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#hermitianObservable.
    def visitHermitianObservable(self, ctx: BraketPragmasParser.HermitianObservableContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#observableResultTypeName.
    def visitObservableResultTypeName(
        self, ctx: BraketPragmasParser.ObservableResultTypeNameContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#standardObservableName.
    def visitStandardObservableName(self, ctx: BraketPragmasParser.StandardObservableNameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#complexOneValue.
    def visitComplexOneValue(self, ctx: BraketPragmasParser.ComplexOneValueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#complexTwoValues.
    def visitComplexTwoValues(self, ctx: BraketPragmasParser.ComplexTwoValuesContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#braketNoisePragma.
    def visitBraketNoisePragma(self, ctx: BraketPragmasParser.BraketNoisePragmaContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#Noise.
    def visitNoise(self, ctx: BraketPragmasParser.NoiseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#Kraus.
    def visitKraus(self, ctx: BraketPragmasParser.KrausContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#matrices.
    def visitMatrices(self, ctx: BraketPragmasParser.MatricesContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#probabilities.
    def visitProbabilities(self, ctx: BraketPragmasParser.ProbabilitiesContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#noiseInstructionName.
    def visitNoiseInstructionName(self, ctx: BraketPragmasParser.NoiseInstructionNameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#program.
    def visitProgram(self, ctx: BraketPragmasParser.ProgramContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#version.
    def visitVersion(self, ctx: BraketPragmasParser.VersionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#statement.
    def visitStatement(self, ctx: BraketPragmasParser.StatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#annotation.
    def visitAnnotation(self, ctx: BraketPragmasParser.AnnotationContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#scope.
    def visitScope(self, ctx: BraketPragmasParser.ScopeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#pragma.
    def visitPragma(self, ctx: BraketPragmasParser.PragmaContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#statementOrScope.
    def visitStatementOrScope(self, ctx: BraketPragmasParser.StatementOrScopeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#calibrationGrammarStatement.
    def visitCalibrationGrammarStatement(
        self, ctx: BraketPragmasParser.CalibrationGrammarStatementContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#includeStatement.
    def visitIncludeStatement(self, ctx: BraketPragmasParser.IncludeStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#breakStatement.
    def visitBreakStatement(self, ctx: BraketPragmasParser.BreakStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#continueStatement.
    def visitContinueStatement(self, ctx: BraketPragmasParser.ContinueStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#endStatement.
    def visitEndStatement(self, ctx: BraketPragmasParser.EndStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#forStatement.
    def visitForStatement(self, ctx: BraketPragmasParser.ForStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#ifStatement.
    def visitIfStatement(self, ctx: BraketPragmasParser.IfStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#returnStatement.
    def visitReturnStatement(self, ctx: BraketPragmasParser.ReturnStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#whileStatement.
    def visitWhileStatement(self, ctx: BraketPragmasParser.WhileStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#barrierStatement.
    def visitBarrierStatement(self, ctx: BraketPragmasParser.BarrierStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#boxStatement.
    def visitBoxStatement(self, ctx: BraketPragmasParser.BoxStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#delayStatement.
    def visitDelayStatement(self, ctx: BraketPragmasParser.DelayStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#gateCallStatement.
    def visitGateCallStatement(self, ctx: BraketPragmasParser.GateCallStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#measureArrowAssignmentStatement.
    def visitMeasureArrowAssignmentStatement(
        self, ctx: BraketPragmasParser.MeasureArrowAssignmentStatementContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#resetStatement.
    def visitResetStatement(self, ctx: BraketPragmasParser.ResetStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#aliasDeclarationStatement.
    def visitAliasDeclarationStatement(
        self, ctx: BraketPragmasParser.AliasDeclarationStatementContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#classicalDeclarationStatement.
    def visitClassicalDeclarationStatement(
        self, ctx: BraketPragmasParser.ClassicalDeclarationStatementContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#constDeclarationStatement.
    def visitConstDeclarationStatement(
        self, ctx: BraketPragmasParser.ConstDeclarationStatementContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#ioDeclarationStatement.
    def visitIoDeclarationStatement(self, ctx: BraketPragmasParser.IoDeclarationStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#oldStyleDeclarationStatement.
    def visitOldStyleDeclarationStatement(
        self, ctx: BraketPragmasParser.OldStyleDeclarationStatementContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#quantumDeclarationStatement.
    def visitQuantumDeclarationStatement(
        self, ctx: BraketPragmasParser.QuantumDeclarationStatementContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#defStatement.
    def visitDefStatement(self, ctx: BraketPragmasParser.DefStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#externStatement.
    def visitExternStatement(self, ctx: BraketPragmasParser.ExternStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#gateStatement.
    def visitGateStatement(self, ctx: BraketPragmasParser.GateStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#assignmentStatement.
    def visitAssignmentStatement(self, ctx: BraketPragmasParser.AssignmentStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#expressionStatement.
    def visitExpressionStatement(self, ctx: BraketPragmasParser.ExpressionStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#defcalStatement.
    def visitDefcalStatement(self, ctx: BraketPragmasParser.DefcalStatementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#bitwiseXorExpression.
    def visitBitwiseXorExpression(self, ctx: BraketPragmasParser.BitwiseXorExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#additiveExpression.
    def visitAdditiveExpression(self, ctx: BraketPragmasParser.AdditiveExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#durationofExpression.
    def visitDurationofExpression(self, ctx: BraketPragmasParser.DurationofExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#parenthesisExpression.
    def visitParenthesisExpression(self, ctx: BraketPragmasParser.ParenthesisExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#comparisonExpression.
    def visitComparisonExpression(self, ctx: BraketPragmasParser.ComparisonExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#multiplicativeExpression.
    def visitMultiplicativeExpression(
        self, ctx: BraketPragmasParser.MultiplicativeExpressionContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#logicalOrExpression.
    def visitLogicalOrExpression(self, ctx: BraketPragmasParser.LogicalOrExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#castExpression.
    def visitCastExpression(self, ctx: BraketPragmasParser.CastExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#powerExpression.
    def visitPowerExpression(self, ctx: BraketPragmasParser.PowerExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#bitwiseOrExpression.
    def visitBitwiseOrExpression(self, ctx: BraketPragmasParser.BitwiseOrExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#callExpression.
    def visitCallExpression(self, ctx: BraketPragmasParser.CallExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#bitshiftExpression.
    def visitBitshiftExpression(self, ctx: BraketPragmasParser.BitshiftExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#bitwiseAndExpression.
    def visitBitwiseAndExpression(self, ctx: BraketPragmasParser.BitwiseAndExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#equalityExpression.
    def visitEqualityExpression(self, ctx: BraketPragmasParser.EqualityExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#logicalAndExpression.
    def visitLogicalAndExpression(self, ctx: BraketPragmasParser.LogicalAndExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#indexExpression.
    def visitIndexExpression(self, ctx: BraketPragmasParser.IndexExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#unaryExpression.
    def visitUnaryExpression(self, ctx: BraketPragmasParser.UnaryExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#literalExpression.
    def visitLiteralExpression(self, ctx: BraketPragmasParser.LiteralExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#aliasExpression.
    def visitAliasExpression(self, ctx: BraketPragmasParser.AliasExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#declarationExpression.
    def visitDeclarationExpression(self, ctx: BraketPragmasParser.DeclarationExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#measureExpression.
    def visitMeasureExpression(self, ctx: BraketPragmasParser.MeasureExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#rangeExpression.
    def visitRangeExpression(self, ctx: BraketPragmasParser.RangeExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#setExpression.
    def visitSetExpression(self, ctx: BraketPragmasParser.SetExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#arrayLiteral.
    def visitArrayLiteral(self, ctx: BraketPragmasParser.ArrayLiteralContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#indexOperator.
    def visitIndexOperator(self, ctx: BraketPragmasParser.IndexOperatorContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#indexedIdentifier.
    def visitIndexedIdentifier(self, ctx: BraketPragmasParser.IndexedIdentifierContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#returnSignature.
    def visitReturnSignature(self, ctx: BraketPragmasParser.ReturnSignatureContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#gateModifier.
    def visitGateModifier(self, ctx: BraketPragmasParser.GateModifierContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#scalarType.
    def visitScalarType(self, ctx: BraketPragmasParser.ScalarTypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#qubitType.
    def visitQubitType(self, ctx: BraketPragmasParser.QubitTypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#arrayType.
    def visitArrayType(self, ctx: BraketPragmasParser.ArrayTypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#arrayReferenceType.
    def visitArrayReferenceType(self, ctx: BraketPragmasParser.ArrayReferenceTypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#designator.
    def visitDesignator(self, ctx: BraketPragmasParser.DesignatorContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#gateOperand.
    def visitGateOperand(self, ctx: BraketPragmasParser.GateOperandContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#externArgument.
    def visitExternArgument(self, ctx: BraketPragmasParser.ExternArgumentContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#defcalArgument.
    def visitDefcalArgument(self, ctx: BraketPragmasParser.DefcalArgumentContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#argumentDefinition.
    def visitArgumentDefinition(self, ctx: BraketPragmasParser.ArgumentDefinitionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#argumentDefinitionList.
    def visitArgumentDefinitionList(self, ctx: BraketPragmasParser.ArgumentDefinitionListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#expressionList.
    def visitExpressionList(self, ctx: BraketPragmasParser.ExpressionListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#defcalArgumentList.
    def visitDefcalArgumentList(self, ctx: BraketPragmasParser.DefcalArgumentListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#identifierList.
    def visitIdentifierList(self, ctx: BraketPragmasParser.IdentifierListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#gateOperandList.
    def visitGateOperandList(self, ctx: BraketPragmasParser.GateOperandListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by BraketPragmasParser#externArgumentList.
    def visitExternArgumentList(self, ctx: BraketPragmasParser.ExternArgumentListContext):
        return self.visitChildren(ctx)


del BraketPragmasParser



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import sys

from braket.default_simulator.simulator import BaseLocalSimulator
from braket.default_simulator.state_vector_simulation import StateVectorSimulation
from braket.device_schema.simulators import (
    GateModelSimulatorDeviceCapabilities,
    GateModelSimulatorDeviceParameters,
)


class StateVectorSimulator(BaseLocalSimulator):
    DEVICE_ID = "braket_sv"

    def initialize_simulation(self, **kwargs) -> StateVectorSimulation:
        """
        Initialize state vector simulation.

        Args:
            `**kwargs`: qubit_count, shots, batch_size

        Returns:
            StateVectorSimulation: Initialized simulation.
        """
        qubit_count = kwargs.get("qubit_count")
        shots = kwargs.get("shots")
        batch_size = kwargs.get("batch_size")
        return StateVectorSimulation(qubit_count, shots, batch_size)

    @property
    def properties(self) -> GateModelSimulatorDeviceCapabilities:
        """
        Device properties for the StateVectorSimulator.

        Returns:
            GateModelSimulatorDeviceCapabilities: Device capabilities for this simulator.
        """
        observables = ["x", "y", "z", "h", "i", "hermitian"]
        max_shots = sys.maxsize
        qubit_count = 26
        return GateModelSimulatorDeviceCapabilities.parse_obj(
            {
                "service": {
                    "executionWindows": [
                        {
                            "executionDay": "Everyday",
                            "windowStartHour": "00:00",
                            "windowEndHour": "23:59:59",
                        }
                    ],
                    "shotsRange": [0, max_shots],
                },
                "action": {
                    "braket.ir.jaqcd.program": {
                        "actionType": "braket.ir.jaqcd.program",
                        "version": ["1"],
                        "supportedOperations": [
                            "ccnot",
                            "cnot",
                            "cphaseshift",
                            "cphaseshift00",
                            "cphaseshift01",
                            "cphaseshift10",
                            "cswap",
                            "cv",
                            "cy",
                            "cz",
                            "ecr",
                            "h",
                            "i",
                            "iswap",
                            "pswap",
                            "phaseshift",
                            "rx",
                            "ry",
                            "rz",
                            "s",
                            "si",
                            "swap",
                            "t",
                            "ti",
                            "unitary",
                            "v",
                            "vi",
                            "x",
                            "xx",
                            "xy",
                            "y",
                            "yy",
                            "z",
                            "zz",
                        ],
                        "supportedResultTypes": [
                            {
                                "name": "Sample",
                                "observables": observables,
                                "minShots": 1,
                                "maxShots": max_shots,
                            },
                            {
                                "name": "Expectation",
                                "observables": observables,
                                "minShots": 0,
                                "maxShots": max_shots,
                            },
                            {
                                "name": "Variance",
                                "observables": observables,
                                "minShots": 0,
                                "maxShots": max_shots,
                            },
                            {"name": "Probability", "minShots": 0, "maxShots": max_shots},
                            {"name": "StateVector", "minShots": 0, "maxShots": 0},
                            {"name": "DensityMatrix", "minShots": 0, "maxShots": 0},
                            {"name": "Amplitude", "minShots": 0, "maxShots": 0},
                        ],
                    },
                    "braket.ir.openqasm.program": {
                        "actionType": "braket.ir.openqasm.program",
                        "version": ["1"],
                        "supportedOperations": [
                            # OpenQASM primitives
                            "U",
                            "GPhase",
                            # builtin Braket gates
                            "ccnot",
                            "cnot",
                            "cphaseshift",
                            "cphaseshift00",
                            "cphaseshift01",
                            "cphaseshift10",
                            "cswap",
                            "cv",
                            "cy",
                            "cz",
                            "ecr",
                            "gpi",
                            "gpi2",
                            "h",
                            "i",
                            "iswap",
                            "ms",
                            "pswap",
                            "phaseshift",
                            "prx",
                            "rx",
                            "ry",
                            "rz",
                            "s",
                            "si",
                            "swap",
                            "t",
                            "ti",
                            "unitary",
                            "v",
                            "vi",
                            "x",
                            "xx",
                            "xy",
                            "y",
                            "yy",
                            "z",
                            "zz",
                        ],
                        "supportedModifiers": [
                            {
                                "name": "ctrl",
                            },
                            {
                                "name": "negctrl",
                            },
                            {
                                "name": "pow",
                                "exponent_types": ["int", "float"],
                            },
                            {
                                "name": "inv",
                            },
                        ],
                        "supportedPragmas": [
                            "braket_unitary_matrix",
                            "braket_result_type_state_vector",
                            "braket_result_type_density_matrix",
                            "braket_result_type_sample",
                            "braket_result_type_expectation",
                            "braket_result_type_variance",
                            "braket_result_type_probability",
                            "braket_result_type_amplitude",
                        ],
                        "forbiddenPragmas": [
                            "braket_noise_amplitude_damping",
                            "braket_noise_bit_flip",
                            "braket_noise_depolarizing",
                            "braket_noise_kraus",
                            "braket_noise_pauli_channel",
                            "braket_noise_generalized_amplitude_damping",
                            "braket_noise_phase_flip",
                            "braket_noise_phase_damping",
                            "braket_noise_two_qubit_dephasing",
                            "braket_noise_two_qubit_depolarizing",
                            "braket_result_type_adjoint_gradient",
                        ],
                        "supportedResultTypes": [
                            {
                                "name": "Sample",
                                "observables": observables,
                                "minShots": 1,
                                "maxShots": max_shots,
                            },
                            {
                                "name": "Expectation",
                                "observables": observables,
                                "minShots": 0,
                                "maxShots": max_shots,
                            },
                            {
                                "name": "Variance",
                                "observables": observables,
                                "minShots": 0,
                                "maxShots": max_shots,
                            },
                            {"name": "Probability", "minShots": 0, "maxShots": max_shots},
                            {"name": "StateVector", "minShots": 0, "maxShots": 0},
                            {"name": "DensityMatrix", "minShots": 0, "maxShots": 0},
                            {"name": "Amplitude", "minShots": 0, "maxShots": 0},
                        ],
                        "supportPhysicalQubits": False,
                        "supportsPartialVerbatimBox": False,
                        "requiresContiguousQubitIndices": False,
                        "requiresAllQubitsMeasurement": False,
                        "supportsUnassignedMeasurements": True,
                        "disabledQubitRewiringSupported": False,
                    },
                },
                "paradigm": {"qubitCount": qubit_count},
                "deviceParameters": GateModelSimulatorDeviceParameters.schema(),
            }
        )


DefaultSimulator = StateVectorSimulator



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import uuid
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, Union

import numpy as np

from braket.default_simulator.observables import Hermitian, TensorProduct
from braket.default_simulator.openqasm.circuit import Circuit
from braket.default_simulator.openqasm.interpreter import Interpreter
from braket.default_simulator.openqasm.program_context import AbstractProgramContext, ProgramContext
from braket.default_simulator.operation import Observable, Operation
from braket.default_simulator.operation_helpers import from_braket_instruction
from braket.default_simulator.result_types import (
    ResultType,
    TargetedResultType,
    from_braket_result_type,
)
from braket.default_simulator.simulation import Simulation
from braket.device_schema import DeviceActionType
from braket.ir.jaqcd import Program as JaqcdProgram
from braket.ir.jaqcd.program_v1 import Results
from braket.ir.jaqcd.shared_models import MultiTarget, OptionalMultiTarget
from braket.ir.openqasm import Program as OpenQASMProgram
from braket.ir.openqasm.program_set_v1 import ProgramSet
from braket.simulator import BraketSimulator
from braket.task_result import (
    AdditionalMetadata,
    GateModelTaskResult,
    ResultTypeValue,
    TaskMetadata,
)
from braket.task_result.program_result_v1 import ProgramResult
from braket.task_result.program_set_executable_result_v1 import (
    ProgramSetExecutableResult,
    ProgramSetExecutableResultMetadata,
)
from braket.task_result.program_set_task_metadata_v1 import ProgramMetadata, ProgramSetTaskMetadata
from braket.task_result.program_set_task_result_v1 import ProgramSetTaskResult

_NOISE_INSTRUCTIONS = frozenset(
    instr.lower().replace("_", "")
    for instr in [
        "amplitude_damping",
        "bit_flip",
        "depolarizing",
        "generalized_amplitude_damping",
        "kraus",
        "pauli_channel",
        "two_qubit_pauli_channel",
        "phase_flip",
        "phase_damping",
        "two_qubit_dephasing",
        "two_qubit_depolarizing",
    ]
)


class OpenQASMSimulator(BraketSimulator, ABC):
    """An abstract simulator that runs an OpenQASM 3 program.

    Translation of individual operations and observables from OpenQASM to the desired format
    is handled by implementing the `AbstractProgramContext` interface. This implementation is
    exposed by implementing the `create_program_context` method, which enables the `parse_program`
    method to translate an entire OpenQASM program:

    >>> class MyProgramContext(AbstractProgramContext):
    >>>     def __init__(self):
    >>>         ...
    >>>
    >>>     def add_gate_instruction(self, gate_name: str, target: Tuple[int], ...):
    >>>         ...
    >>>
    >>>     # Implement other MyProgramContext interface methods
    >>>
    >>> class MySimulator(OpenQASMSimulator):
    >>>     def create_program_context(self) -> AbstractProgramContext:
    >>>         return MyProgramContext()
    >>>
    >>>     # Implement other BraketSimulator interface methods
    >>>
    >>> parsed = MySimulator().parse_program(program)

    To register a simulator so the Amazon Braket SDK recognizes its name,
    the name and class must be added as an entry point for "braket.simulators".
    This is done by adding an entry to entry_points in the simulator package's setup.py:

    >>> entry_points = {
    >>>     "braket.simulators": [
    >>>         "backend_name = <backend_class>"
    >>>     ]
    >>> }
    """

    @abstractmethod
    def create_program_context(self) -> AbstractProgramContext:
        """Creates a new program context to handle translation of OpenQASM into a desired format."""

    def parse_program(self, program: OpenQASMProgram) -> AbstractProgramContext:
        """Parses an OpenQASM program and returns a program context.

        Args:
            program (OpenQASMProgram): The program to parse.

        Returns:
            AbstractProgramContext: The program context after the program has been parsed.
        """
        is_file = program.source.endswith(".qasm")
        interpreter = Interpreter(self.create_program_context())
        return interpreter.run(
            source=program.source,
            inputs=program.inputs,
            is_file=is_file,
        )


class BaseLocalSimulator(OpenQASMSimulator):
    def run(
        self, circuit_ir: Union[OpenQASMProgram, ProgramSet, JaqcdProgram], *args, **kwargs
    ) -> GateModelTaskResult:
        """
        Simulate a program using either OpenQASM or Jaqcd.

        Args:
            circuit_ir (Union[OpenQASMProgram, ProgramSet, JaqcdProgram]): Program
                specification.
            shots (int, optional): The number of shots to simulate. Default is 0, which
                performs a full analytical simulation.
            batch_size (int, optional): The size of the program partitions to contract,
                if applying multiple gates at a time is desired; see `StateVectorSimulation`.
                Must be a positive integer.
                Defaults to 1, which means gates are applied one at a time without any
                optimized contraction.

        Returns:
            GateModelTaskResult: object that represents the result

        Raises:
            ValueError: If result types are not specified in the IR or sample is specified
                as a result type when shots=0. Or, if StateVector and Amplitude result types
                are requested when shots>0.
        """
        if isinstance(circuit_ir, OpenQASMProgram):
            return self.run_openqasm(circuit_ir, *args, **kwargs)
        elif isinstance(circuit_ir, ProgramSet):
            return self.run_program_set(circuit_ir, *args, **kwargs)
        return self.run_jaqcd(circuit_ir, *args, **kwargs)

    def create_program_context(self) -> AbstractProgramContext:
        return ProgramContext()

    @abstractmethod
    def initialize_simulation(self, **kwargs) -> Simulation:
        """Initializes simulation with keyword arguments"""

    def _validate_ir_results_compatibility(
        self, results: list[Results], device_action_type
    ) -> None:
        """
        Validate that requested result types are valid for the simulator.

        Args:
            results (list[Results]): Requested result types.

        Raises:
            TypeError: If any the specified result types are not supported
        """
        if results:
            circuit_result_types_name = [result.__class__.__name__ for result in results]
            supported_result_types = self.properties.action[device_action_type].supportedResultTypes
            supported_result_types_name = [result.name for result in supported_result_types]
            for name in circuit_result_types_name:
                if name not in supported_result_types_name:
                    raise TypeError(
                        f"result type {name} is not supported by {self.__class__.__name__}"
                    )

    @staticmethod
    def _validate_shots_and_ir_results(
        shots: int,
        results: list[Results],
        qubit_count: int,
    ) -> None:
        """
        Validated that requested result types are valid for given shots and qubit count.

        Args:
            shots (int): Shots for the simulation.
            results (list[Results]): Specified result types.
            qubit_count (int): Number of qubits for the simulation.

        Raises:
            ValueError: If any of the requested result types are incompatible with the
                qubit count or number of shots.
        """
        if not shots:
            if not results:
                raise ValueError("Result types must be specified in the IR when shots=0")
            for rt in results:
                if rt.type in ["sample"]:
                    raise ValueError("sample can only be specified when shots>0")
                if rt.type == "amplitude":
                    BaseLocalSimulator._validate_amplitude_states(rt.states, qubit_count)
        elif shots and results:
            for rt in results:
                if rt.type in ["statevector", "amplitude", "densitymatrix"]:
                    raise ValueError(
                        "statevector, amplitude and densitymatrix result "
                        "types not available when shots>0"
                    )

    @staticmethod
    def _validate_amplitude_states(states: list[str], qubit_count: int) -> None:
        """
        Validate states in an amplitude result type are valid.

        Args:
            states (list[str]): List of binary strings representing quantum states.
            qubit_count (int): Number of qubits for the simulation.

        Raises:
            ValueError: If any of the states is not the correct size for the number of qubits.
        """
        for state in states:
            if len(state) != qubit_count:
                raise ValueError(
                    f"Length of state {state} for result type amplitude"
                    f" must be equivalent to number of qubits {qubit_count} in circuit"
                )

    @staticmethod
    def _translate_result_types(results: list[Results]) -> list[ResultType]:
        return [from_braket_result_type(result) for result in results]

    @staticmethod
    def _generate_results(
        results: list[Results],
        result_types: list[ResultType],
        simulation: Simulation,
    ) -> list[ResultTypeValue]:
        return [
            ResultTypeValue.construct(
                type=results[index],
                value=result_types[index].calculate(simulation),
            )
            for index in range(len(results))
        ]

    def _create_results_obj(
        self,
        results: list[dict[str, Any]],
        openqasm_ir: OpenQASMProgram,
        simulation: Simulation,
        measured_qubits: list[int] = None,
        mapped_measured_qubits: list[int] = None,
    ) -> GateModelTaskResult:
        return GateModelTaskResult.construct(
            taskMetadata=TaskMetadata(
                id=str(uuid.uuid4()),
                shots=simulation.shots,
                deviceId=self.DEVICE_ID,
            ),
            additionalMetadata=AdditionalMetadata(
                action=openqasm_ir,
            ),
            resultTypes=results,
            measurements=self._formatted_measurements(simulation, mapped_measured_qubits),
            measuredQubits=(measured_qubits or list(range(simulation.qubit_count))),
        )

    @staticmethod
    def _get_qubits_referenced(operations: list[Operation]) -> set[int]:
        return {target for operation in operations for target in operation.targets}

    @staticmethod
    def _validate_result_types_qubits_exist(
        targeted_result_types: list[TargetedResultType], qubit_count: int
    ) -> None:
        for result_type in targeted_result_types:
            targets = result_type.targets
            if targets and max(targets) >= qubit_count:
                raise ValueError(
                    f"Result type ({result_type.__class__.__name__})"
                    f" references invalid qubits {targets}"
                )

    def _validate_ir_instructions_compatibility(
        self,
        circuit_ir: Union[JaqcdProgram, Circuit],
        device_action_type: DeviceActionType,
    ) -> None:
        """
        Validate that requested IR instructions are valid for the simulator.

        Args:
            circuit_ir (Union[JaqcdProgram, Circuit]): IR for the simulator.

        Raises:
            TypeError: If any the specified result types are not supported
        """
        circuit_instruction_names = [
            instr.__class__.__name__.lower().replace("_", "") for instr in circuit_ir.instructions
        ]
        supported_instructions = frozenset(
            op.lower().replace("_", "")
            for op in self.properties.action[device_action_type].supportedOperations
        )
        no_noise = True
        for name in circuit_instruction_names:
            if name in _NOISE_INSTRUCTIONS:
                no_noise = False
                if name not in supported_instructions:
                    raise TypeError(
                        "Noise instructions are not supported by the state vector simulator "
                        "(by default). You need to use the density matrix simulator: "
                        'LocalSimulator("braket_dm").'
                    )
        if no_noise and _NOISE_INSTRUCTIONS.intersection(supported_instructions):
            warnings.warn(
                "You are running a noise-free circuit on the density matrix simulator. "
                "Consider running this circuit on the state vector simulator: "
                'LocalSimulator("default") for a better user experience.'
            )

    def _validate_input_provided(self, circuit: Circuit) -> None:
        """
        Validate that requested circuit has all input parameters provided.

        Args:
            circuit (Circuit): IR for the simulator.

        Raises:
            NameError: If any the specified input parameters are not provided
        """
        for instruction in circuit.instructions:
            possible_parameters = "_angle", "_angle_1", "_angle_2"
            for parameter_name in possible_parameters:
                param = getattr(instruction, parameter_name, None)
                if param is not None:
                    try:
                        float(param)
                    except TypeError:
                        missing_input = param.free_symbols.pop()
                        raise NameError(f"Missing input variable '{missing_input}'.")

    @staticmethod
    def _tensor_product_index_dict(
        observable: TensorProduct, func: Callable[[Observable], Any]
    ) -> dict[int, Any]:
        obj_dict = {}
        i = 0
        factors = list(observable.factors)
        total = len(factors[0].measured_qubits)
        while factors:
            if i >= total:
                factors.pop(0)
                if factors:
                    total += len(factors[0].measured_qubits)
            if factors:
                obj_dict[i] = func(factors[0])
            i += 1
        return obj_dict

    @staticmethod
    def _observable_hash(observable: Observable) -> Union[str, dict[int, str]]:
        if isinstance(observable, Hermitian):
            return str(hash(str(observable.matrix.tobytes())))
        elif isinstance(observable, TensorProduct):
            # Dict of target index to observable hash
            return BaseLocalSimulator._tensor_product_index_dict(
                observable, BaseLocalSimulator._observable_hash
            )
        else:
            return str(observable.__class__.__name__)

    @staticmethod
    def _map_circuit_to_contiguous_qubits(circuit: Union[Circuit, JaqcdProgram]) -> dict[int, int]:
        """
        Maps the qubits in operations and result types to contiguous qubits.

        Args:
            circuit (Union[Circuit, JaqcdProgram]): The circuit containing the operations and
            result types.

        Returns:
            dict[int, int]: Map of qubit index to corresponding contiguous index
        """
        circuit_qubit_set = BaseLocalSimulator._get_circuit_qubit_set(circuit)
        qubit_map = BaseLocalSimulator._contiguous_qubit_mapping(circuit_qubit_set)
        BaseLocalSimulator._map_circuit_qubits(circuit, qubit_map)
        return qubit_map

    @staticmethod
    def _get_circuit_qubit_set(circuit: Union[Circuit, JaqcdProgram]) -> set[int]:
        """
        Returns the set of qubits used in the given circuit.

        Args:
            circuit (Union[Circuit, JaqcdProgram]): The circuit from which to extract the qubit set.

        Returns:
            set[int]: The set of qubits used in the circuit.
        """
        if isinstance(circuit, Circuit):
            return circuit.qubit_set
        else:
            operations = [
                from_braket_instruction(instruction) for instruction in circuit.instructions
            ]
            if circuit.basis_rotation_instructions:
                operations.extend(
                    from_braket_instruction(instruction)
                    for instruction in circuit.basis_rotation_instructions
                )
            return BaseLocalSimulator._get_qubits_referenced(operations)

    @staticmethod
    def _map_circuit_qubits(circuit: Union[Circuit, JaqcdProgram], qubit_map: dict[int, int]):
        """
        Maps the qubits in operations and result types to contiguous qubits.

        Args:
            circuit (Circuit): The circuit containing the operations and result types.
            qubit_map (dict[int, int]): The mapping from qubits to their contiguous indices.

        Returns:
            Circuit: The circuit with qubits in operations and result types mapped
            to contiguous qubits.
        """
        if isinstance(circuit, Circuit):
            BaseLocalSimulator._map_circuit_instructions(circuit, qubit_map)
            BaseLocalSimulator._map_circuit_results(circuit, qubit_map)
        else:
            BaseLocalSimulator._map_jaqcd_instructions(circuit, qubit_map)
        return circuit

    @staticmethod
    def _map_circuit_instructions(circuit: Circuit, qubit_map: dict):
        """
        Maps the targets of each instruction in the circuit to the corresponding qubits in the
        qubit_map.

        Args:
            circuit (Circuit): The circuit containing the instructions.
            qubit_map (dict): A dictionary mapping original qubits to new qubits.
        """
        for ins in circuit.instructions:
            ins._targets = tuple([qubit_map[q] for q in ins.targets])

    @staticmethod
    def _map_circuit_results(circuit: Circuit, qubit_map: dict):
        """
        Maps the targets of each result in the circuit to the corresponding qubits in the qubit_map.

        Args:
            circuit (Circuit): The circuit containing the results.
            qubit_map (dict): A dictionary mapping original qubits to new qubits.
        """
        for result in circuit.results:
            if isinstance(result, (MultiTarget, OptionalMultiTarget)) and result.targets:
                result.targets = [qubit_map[q] for q in result.targets]

    @staticmethod
    def _map_jaqcd_instructions(circuit: JaqcdProgram, qubit_map: dict):
        """
        Maps the attributes of each instruction in the JaqcdProgram to the corresponding qubits in
        the qubit_map.

        Args:
            circuit (JaqcdProgram): The JaqcdProgram containing the instructions.
            qubit_map (dict): A dictionary mapping original qubits to new qubits.
        """
        for ins in circuit.instructions:
            BaseLocalSimulator._map_instruction_attributes(ins, qubit_map)

        if hasattr(circuit, "results") and circuit.results:
            for ins in circuit.results:
                BaseLocalSimulator._map_instruction_attributes(ins, qubit_map)

        if circuit.basis_rotation_instructions:
            for ins in circuit.basis_rotation_instructions:
                ins.target = qubit_map[ins.target]

    @staticmethod
    def _map_instruction_attributes(instruction, qubit_map: dict):
        """
        Maps the qubit attributes of an instruction from JaqcdProgram to the corresponding
        qubits in the qubit_map.

        Args:
            instruction: The Jaqcd instruction whose qubit attributes need to be mapped.
            qubit_map (dict): A dictionary mapping original qubits to new qubits.
        """
        if hasattr(instruction, "control"):
            instruction.control = qubit_map.get(instruction.control, instruction.control)

        if hasattr(instruction, "controls") and instruction.controls:
            instruction.controls = [qubit_map.get(q, q) for q in instruction.controls]

        if hasattr(instruction, "target"):
            instruction.target = qubit_map.get(instruction.target, instruction.target)

        if hasattr(instruction, "targets") and instruction.targets:
            instruction.targets = [qubit_map.get(q, q) for q in instruction.targets]

    @staticmethod
    def _contiguous_qubit_mapping(qubit_set: set[int]) -> dict[int, int]:
        """
        Maping of qubits to contiguous integers. The qubit mapping may be discontiguous or
        contiguous.

        Args:
            qubit_set (set[int]): List of qubits to be mapped.

        Returns:
            dict[int, int]: Dictionary where keys are qubits and values are contiguous integers.
        """
        return {q: i for i, q in enumerate(sorted(qubit_set))}

    @staticmethod
    def _formatted_measurements(
        simulation: Simulation, measured_qubits: Union[list[int], None] = None
    ) -> list[list[str]]:
        """Retrieves formatted measurements obtained from the specified simulation.

        Args:
            simulation (Simulation): Simulation to use for obtaining the measurements.
            measured_qubits (list[int] | None): The qubits that were measured.

        Returns:
            list[list[str]]: List containing the measurements, where each measurement consists
            of a list of measured values of qubits.
        """
        # Get the full measurements
        measurements = [
            list("{number:0{width}b}".format(number=sample, width=simulation.qubit_count))
            for sample in simulation.retrieve_samples()
        ]
        #  Gets the subset of measurements from the full measurements
        if measured_qubits is not None and measured_qubits != []:
            measured_qubits = np.array(measured_qubits)
            in_circuit_mask = measured_qubits < simulation.qubit_count
            measured_qubits_in_circuit = measured_qubits[in_circuit_mask]
            measured_qubits_not_in_circuit = measured_qubits[~in_circuit_mask]

            measurements_array = np.array(measurements)
            selected_measurements = measurements_array[:, measured_qubits_in_circuit]
            measurements = np.pad(
                selected_measurements, ((0, 0), (0, len(measured_qubits_not_in_circuit)))
            ).tolist()
        return measurements

    def _run_single_program_in_program_set(
        self,
        program: OpenQASMProgram,
        shots: int = 0,
        *,
        batch_size: int = 1,
    ) -> ProgramResult:
        """Executes the program specified by the supplied `program` on the simulator.

        Args:
            program_set_ir (OpenQASMProgram): ir representation of the program.
            shots (int): The number of times to run each executable in the program.
            batch_size (int): The size of the circuit partitions to contract for each executable
                in the program, if applying multiple gates at a time is desired; see
                `StateVectorSimulation`. Must be a positive integer. Defaults to 1, which means
                gates are applied one at a time without any contraction.

        Returns:
            ProgramResult: Result of the program simulation.
        """
        if program.inputs:
            # Unpack inputs. For example, it unpacks {"alpha": [1, 2], "beta": [0.1, 0.2]} from a
            # program to [{"alpha": 1, "beta": 0.1}, {"alpha": 2, "beta": 0.2}] for executables.
            unpacked_inputs = [
                dict(zip(program.inputs.keys(), values)) for values in zip(*program.inputs.values())
            ]
            executables = [
                OpenQASMProgram(source=program.source, inputs=input_param)
                for input_param in unpacked_inputs
            ]
        else:
            executables = [OpenQASMProgram(source=program.source)]

        executable_results = []
        for input_index, executable in enumerate(executables):
            executable_raw_result = self.run_openqasm(
                executable,
                shots=shots,
                batch_size=batch_size,
            )
            executable_results.append(
                ProgramSetExecutableResult(
                    inputsIndex=input_index,
                    measurements=executable_raw_result.measurements,
                    measurementProbabilities=executable_raw_result.measurementProbabilities,
                    measuredQubits=executable_raw_result.measuredQubits,
                )
            )

        program_result = ProgramResult(
            executableResults=executable_results,
            source=program,
            additionalMetadata=AdditionalMetadata(),
        )
        program_metadata = ProgramMetadata(
            executables=[ProgramSetExecutableResultMetadata()] * len(executables)
        )
        return program_result, program_metadata

    def run_program_set(
        self,
        program_set: ProgramSet,
        shots: int = 0,
        *,
        batch_size: int = 1,
    ) -> ProgramSetTaskResult:
        """Executes the program set specified by the supplied `program_set_ir` on the simulator.

        Args:
            program_set (ProgramSet): IR representation of the program set.
            shots (int): The number of times to run each executable in the program set.
            batch_size (int): The size of the circuit partitions to contract for each executable
                in the program set, if applying multiple gates at a time is desired; see
                `StateVectorSimulation`. Must be a positive integer. Defaults to 1, which means
                gates are applied one at a time without any contraction.

        Returns:
            ProgramSetTaskResult: Result of the program set simulation.
        """

        if shots == 0:
            raise ValueError(
                "Shots must not be zero. Result types are not supported with program set."
            )
        shots_per_executable, remainder = divmod(shots, program_set.num_executables)
        if remainder:
            raise ValueError("Total shots must be divisible by number of executables.")

        program_results_metadata = [
            self._run_single_program_in_program_set(
                program, shots=shots_per_executable, batch_size=batch_size
            )
            for program in program_set.programs
        ]

        program_results, program_metadata = zip(*program_results_metadata)
        return ProgramSetTaskResult(
            programResults=program_results,
            taskMetadata=ProgramSetTaskMetadata(
                id=str(uuid.uuid4()),
                deviceId=self.DEVICE_ID,
                requestedShots=shots,
                successfulShots=shots,
                programMetadata=program_metadata,
                totalFailedExecutables=0,
            ),
        )

    def run_openqasm(
        self,
        openqasm_ir: OpenQASMProgram,
        shots: int = 0,
        *,
        batch_size: int = 1,
    ) -> GateModelTaskResult:
        """Executes the circuit specified by the supplied `circuit_ir` on the simulator.

        Args:
            openqasm_ir (Program): ir representation of a braket circuit specifying the
                instructions to execute.
            shots (int): The number of times to run the circuit.
            batch_size (int): The size of the circuit partitions to contract,
                if applying multiple gates at a time is desired; see `StateVectorSimulation`.
                Must be a positive integer.
                Defaults to 1, which means gates are applied one at a time without any
                optimized contraction.
        Returns:
            GateModelTaskResult: object that represents the result

        Raises:
            ValueError: If result types are not specified in the IR or sample is specified
                as a result type when shots=0. Or, if StateVector and Amplitude result types
                are requested when shots>0.
        """
        circuit = self.parse_program(openqasm_ir).circuit
        qubit_map = BaseLocalSimulator._map_circuit_to_contiguous_qubits(circuit)
        qubit_count = circuit.num_qubits
        measured_qubits = circuit.measured_qubits
        mapped_measured_qubits = (
            [qubit_map[q] for q in measured_qubits] if measured_qubits else None
        )

        self._validate_ir_results_compatibility(
            circuit.results,
            device_action_type=DeviceActionType.OPENQASM,
        )
        self._validate_ir_instructions_compatibility(
            circuit,
            device_action_type=DeviceActionType.OPENQASM,
        )
        self._validate_input_provided(circuit)
        BaseLocalSimulator._validate_shots_and_ir_results(shots, circuit.results, qubit_count)

        results = circuit.results

        simulation = self.initialize_simulation(
            qubit_count=qubit_count, shots=shots, batch_size=batch_size
        )
        operations = circuit.instructions
        simulation.evolve(operations)

        if not shots:
            result_types = BaseLocalSimulator._translate_result_types(results)
            BaseLocalSimulator._validate_result_types_qubits_exist(
                [
                    result_type
                    for result_type in result_types
                    if isinstance(result_type, TargetedResultType)
                ],
                qubit_count,
            )
            results = BaseLocalSimulator._generate_results(
                circuit.results,
                result_types,
                simulation,
            )
        else:
            simulation.evolve(circuit.basis_rotation_instructions)

        return self._create_results_obj(
            results, openqasm_ir, simulation, measured_qubits, mapped_measured_qubits
        )

    def run_jaqcd(
        self,
        circuit_ir: JaqcdProgram,
        qubit_count: Any = None,
        shots: int = 0,
        *,
        batch_size: int = 1,
    ) -> GateModelTaskResult:
        """Executes the circuit specified by the supplied `circuit_ir` on the simulator.

        Args:
            circuit_ir (Program): ir representation of a braket circuit specifying the
                instructions to execute.
            qubit_count (Any): Unused parameter; in signature for backwards-compatibility
            shots (int): The number of times to run the circuit.
            batch_size (int): The size of the circuit partitions to contract,
                if applying multiple gates at a time is desired; see `StateVectorSimulation`.
                Must be a positive integer.
                Defaults to 1, which means gates are applied one at a time without any
                optimized contraction.
        Returns:
            GateModelTaskResult: object that represents the result

        Raises:
            ValueError: If result types are not specified in the IR or sample is specified
                as a result type when shots=0. Or, if StateVector and Amplitude result types
                are requested when shots>0.
        """
        if qubit_count is not None:
            warnings.warn(
                f"qubit_count is deprecated for {type(self).__name__} and can be set to None"
            )
        self._validate_ir_results_compatibility(
            circuit_ir.results,
            device_action_type=DeviceActionType.JAQCD,
        )
        self._validate_ir_instructions_compatibility(
            circuit_ir,
            device_action_type=DeviceActionType.JAQCD,
        )
        qubit_map = BaseLocalSimulator._map_circuit_to_contiguous_qubits(circuit_ir)
        qubit_count = len(qubit_map)
        BaseLocalSimulator._validate_shots_and_ir_results(shots, circuit_ir.results, qubit_count)

        operations = [
            from_braket_instruction(instruction) for instruction in circuit_ir.instructions
        ]

        if shots > 0 and circuit_ir.basis_rotation_instructions:
            operations.extend(
                from_braket_instruction(instruction)
                for instruction in circuit_ir.basis_rotation_instructions
            )

        simulation = self.initialize_simulation(
            qubit_count=qubit_count, shots=shots, batch_size=batch_size
        )
        simulation.evolve(operations)

        results = []

        if not shots and circuit_ir.results:
            result_types = BaseLocalSimulator._translate_result_types(circuit_ir.results)
            BaseLocalSimulator._validate_result_types_qubits_exist(
                [
                    result_type
                    for result_type in result_types
                    if isinstance(result_type, TargetedResultType)
                ],
                qubit_count,
            )
            results = self._generate_results(
                circuit_ir.results,
                result_types,
                simulation,
            )

        return self._create_results_obj(results, circuit_ir, simulation)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from decimal import Decimal

from pydantic import PositiveInt
from pydantic.v1.main import BaseModel


class CapabilitiesConstants(BaseModel):
    DIMENSIONS: PositiveInt = 2
    BOUNDING_BOX_SIZE_X: Decimal
    BOUNDING_BOX_SIZE_Y: Decimal
    MAX_TIME: Decimal
    MIN_DISTANCE: Decimal
    GLOBAL_AMPLITUDE_VALUE_MIN: Decimal
    GLOBAL_AMPLITUDE_VALUE_MAX: Decimal
    GLOBAL_DETUNING_VALUE_MIN: Decimal
    GLOBAL_DETUNING_VALUE_MAX: Decimal

    LOCAL_MAGNITUDE_SEQUENCE_VALUE_MIN: Decimal
    LOCAL_MAGNITUDE_SEQUENCE_VALUE_MAX: Decimal

    MAGNITUDE_PATTERN_VALUE_MIN: Decimal
    MAGNITUDE_PATTERN_VALUE_MAX: Decimal
    MAX_NET_DETUNING: Decimal
